import functools
from flask import url_for, current_app
from bento_meta.mdb import SearchableMDB
from bento_meta.model import Model
from bento_meta.object_map import ObjectMap


plural = {
    "node": "nodes",
    "edge": "edges",
    "relationship": "relationships",
    "property": "properties",
    "tag": "tags",
    "origin": "origins",
    "concept": "concepts",
    "predicate": "predicates",
    "value_set": "value_sets",
    "term": "terms",
    }


class mdb():
    """Read functionality for driving STS UI. Mixins mdb_update and mdb_tags
could be used here for write and tag functionality."""
    mdb_ = None
    term_values = None
    def __init__(self):
        if mdb.mdb_ is None:
            mdb.mdb_ = SearchableMDB(current_app.config["NEO4J_MDB_URI"],
                                     user=current_app.config["NEO4J_MDB_USER"],
                                     password=current_app.config["NEO4J_MDB_PASS"])
        self.mdb = mdb.mdb_

    def close(self):
        self.mdb.close()


    def get_list_of_models(self):
        models = self.mdb.get_model_nodes()
        ret = {}
        if models:
            for mdl in models:
                h = mdl["m"]["handle"]
                if ret.get(h):
                    ret[h]["version"].append(mdl["m"]["version"])
                else:
                    ret[h] = mdl["m"]
                    if not ret[h].get("name"):
                        ret[h]["name"] = ret[h]["handle"]
                    ret[h]["version"] = [mdl["m"]["version"]]
        return [x for x in ret.values()]
    def get_model_by_name(self, name):
        model_nodes = self.mdb.get_model_nodes(model=name)
        models = []
        for node in model_nodes:
            model = Model(handle=name)
            model.version = node["m"].get("version")
            model.repository = node["m"].get("repository")
            models.append(model)
        return models


    # ####################################################################### #
    # NODES
    # ####################################################################### #
    def get_node_by_id(self, nid, model=None):
        props_result = self.mdb.get_node_and_props_by_node_id(nid)
        edges_result = self.mdb.get_node_edges_by_node_id(nid)
        to_nodes = {}
        from_nodes = {}
        if not props_result and not edges_result:
            return {}
        result = {"id": nid,
                  "handle": props_result[0]["handle"],
                  "model": props_result[0]["model"],
                  "version": props_result[0].get("version"),
                  "has_properties": [],
                  "has_relationship_to_nodes": [],
                  "has_relationship_from_nodes": []}
        if (props_result):
            for p in props_result[0]["props"]:
                result["has_properties"].append({
                    "id": p["nanoid"],
                    "handle": p["handle"],
                    "model": p["model"],
                    "version": p.get("version"),
                    "type": "property",
                    "link": url_for("main.entities", entities='properties',
                                    id=p["nanoid"]),
                    "value_domain": p["value_domain"]
                    })
        if (edges_result):
            for e in edges_result:
                node = e["far_node"]
                if (node):
                    rln = e["rln"]
                    add_to = None
                    if e["far_type"] == "has_src":
                        add_to = from_nodes
                    elif e["far_type"] == "has_dst":
                        add_to = to_nodes
                    else:
                        RuntimeError("What?")
                    add_to[node["nanoid"]] = {
                        "id": node["nanoid"],
                        "handle": node["handle"],
                        "model": node["model"],
                        "version": node.get("version"),
                        "link": url_for("main.entities", entities='nodes',
                                        id=node["nanoid"]),
                        "type": "node",
                        "relationship": {
                            "id": rln["nanoid"],
                            "handle": rln["handle"],
                            "model": rln["model"],
                            "version": rln.get("version"),
                            "link": "/relationships/{}".format(rln["nanoid"]),
                            "type": "relationship"
                        }
                    }
            result["has_relationship_to_nodes"].extend(to_nodes.values())
            result["has_relationship_from_nodes"].extend(from_nodes.values())
        return result

    # ----------------------------------------------------------------------- #

    def get_list_of_nodes(self, model=None, version=None):
        """
        In [3]: m.get_list_of_nodes()
        Out[3]:
        [{'yXWr0Y': 'study_site'},
         {'N0Qx7Z': 'off_study'},
         {'nUoHJH': 'diagnosis
        """
        if model == 'ALL':
            model = None
            version = None
        elif version == 'ALL':
            version = '*'
        result = self.mdb.get_nodes_by_model(model, version)
        if result:
            return [(x["nanoid"], x["handle"],
                     x["model"], x["version"]) for x in result]
        else:
            return []

    # ####################################################################### #
    # PROPERTIES
    # ####################################################################### #

    def get_list_of_properties(self, model=None, version=None):
        if model == 'ALL':
            model = None
            version = None
        elif version == 'ALL':
            version = '*'
        np_result = self.mdb.get_nodes_and_props_by_model(model, version)
        if not np_result:
            return []
        result = []
        for np in np_result:
            result.extend([{"prop_id": p["nanoid"], "prop_handle": p["handle"],
                            "node_model": np["model"], "node_handle": np["handle"],
                            "node_id": np["id"], "node_version": np.get("version")}
                           for p in np["props"]])
        return result

    def get_property_by_id(self, pid, model=None):
        p_result = self.mdb.get_prop_node_and_domain_by_prop_id(pid)
        if not p_result:
            return {}
        pr = p_result[0]
        result = {
            "prop": pr["prop"],
            "model": pr["model"],
            "version": pr.get("version"),
            "type": "property",
            "link": "/properties/{}".format(pr["id"]),
            "_for_nodehandle": pr["node"]["handle"],
            "_for_nodeid": pr["node"]["nanoid"]
            }
        if pr["value_set"]:
            result["has_valueset"] = {
                "id": pr["value_set"]["nanoid"],
                "type": "valueset",
                "link": url_for("main.entities", entities='valuesets',
                                id=pr["value_set"]["nanoid"])
                }
        if pr["terms"]:
            result["has_terms"] = [{"id": t["nanoid"], "value": t["value"],
                                    "origin": t["origin_name"] if "origin_name" in t else None,
                                    "type": "term",
                                    "link": url_for("main.entities",
                                                    entities='terms',
                                                    id=t["nanoid"])}
                                   for t in pr["terms"] if 'nanoid' in t]
        return result

    # ####################################################################### #
    # VALUESETS
    # ####################################################################### #

    def get_valueset_by_id(self, vsid, model=None):
        vs_result = self.mdb.get_valueset_by_id(vsid)
        if not vs_result:
            return {}
        result = {
            "id": vs_result[0]["nanoid"],
            "handle": vs_result[0]["handle"],
            "model": model,  # kludge
            "url": vs_result[0]["url"],
            "desc": vs_result[0]["desc"],
            "type": "valueset",
            "link": url_for("main.entities", entities='valuesets',
                            id=vs_result[0]["nanoid"]),
            "_for_propertyhandle": "_for_propertyhandle",  # kludge
            "_for_propertyid": "_for_propertyid"  # kludge
            }
        result["has_property"] = [{"id": x["nanoid"], "handle": x["handle"],
                                   "type": "property", "model": x["model"],
                                   "link": "/properties/{}".format(x["nanoid"])}
                                  for x in vs_result[0].props]
        result["has_terms"] = [{"id": x["nanoid"], "value": x["value"],
                                "origin": x["origin_name"] if "origin_name" in x else None,
                                "type": "term",
                                "link":url_for('main.entities', entities='terms',
                                               id=x["nanoid"])}
                               for x in vs_result[0].terms]
        if model:
            result["has_property"] = [x for x in result["has_property"]
                                      if x["model"] == model]
        # kludge - original assumption is that a valueset has only one
        # associated property, but this is not necessarily true.
        if result["has_property"]:
            result["_for_propertyhandle"] = result["has_property"][0]["handle"]
            result["_for_propertyid"] = result["has_property"][0]["nanoid"]
        return result

    # ----------------------------------------------------------------------- #

    def get_list_of_valuesets(self, model=None, version=None):
        vs_result = self.mdb.get_valuesets_by_model(model, version)
        if not vs_result:
            return []
        # kludge assuming that there is only one property assoc with each
        # valueset:
        return [{"id": x["value_set"]["nanoid"],
                 "handle": x["props"][0]["handle"]}
                for x in vs_result]

    # ####################################################################### #
    # TERMS
    # ####################################################################### #

    def get_term_batch_info(self, nbatches, start=0, ct=None):
        batches = []
        tabnames = []
        if not mdb.term_values:
            try:
                res = self.mdb.get_with_statement(
                    "match (t:term) with t.value as val return val order by val",
                    {})
            except Exception as e:
                raise e
            mdb.term_values = [x['val'] for x in res]
        if ct is None:
            ct = len(self.term_values)
        bsize = ct // nbatches or ct
        for n in range(start, min(start+ct, len(self.term_values)), bsize):
            l = min(n+bsize-1,len(self.term_values)-1)
            batches.append({'first': n, 'first_term': self.term_values[n],
                            'last': l, 'last_term': self.term_values[l]})
            
        # if n+bsize != start+ct:
        #     batches[-1]['last'] = start+ct-1
        #     batches[-1]['last_term'] = self.term_values[start+ct-1]

        for n in range(0, len(batches)):
            v1 = batches[n]['first_term'] or "___"
            v2 = batches[n]['last_term'] or "___"
            tabnames.append("{}-{}".format(v1[0:3],v2[0:3]))

        return (batches, tabnames)

    @functools.lru_cache
    def get_term_batch(start, bsize):
        try:
            res = mdb.mdb_.get_with_statement(
                "match (t:term) with t order by t.value "
                "return t as term skip $start limit $bsize",
                {"start": start, "bsize": bsize}
            )
        except Exception as e:
            raise e
        return res


    def get_term_by_id(self, tid):
        t_result = self.mdb.get_with_statement(
            "match (t:term {nanoid:$id}) "
            "optional match (t)<-[:has_term]-(v:value_set)<-[:has_value_set]-(p:property) "
            "with t as term, t.origin_name as oname,  collect(p) as props "
            "optional match (o:origin) where o.name = oname "
            "return term, o as origin, props ", {"id": tid})
        if not t_result:
            return {}
        result = t_result[0]
        result["props"] = sorted(result["props"],
                                 key=lambda x:(x['handle'],x['model'],x.get('version')))
        s_result = self.mdb.get_with_statement(
            "match (t:term {nanoid:$id})-[:represents]->(c:concept)<-[:represents]-(s:term) "
            "optional match (g:tag {key:'mapping_source'})<-[:has_tag]-(c) "
            "return s.value as value, s.nanoid as id, s.origin_name as origin, collect(g.value) as sources", {"id": tid}
            )
        if s_result:
            result["synonyms"] = sorted(s_result, key=lambda x:x['value'])
        else:
            result["synonyms"] = None
        return result

         
    def get_list_of_terms(self, start=None, end=None):
        t_result = self.mdb.get_with_statement(
            "match (t:term)<-[:has_term]-(:value_set) "
            "with distinct t as t "
            "return t.value as value, t as term", {})
        return t_result

        
    # ####################################################################### #
    # ORIGINS
    # ####################################################################### #

    def get_list_of_origins(self, dummy=None):
        origins = self.mdb.get_origins()
        result = [{x["o"]["nanoid"]:x["o"]["name"]}
                  for x in origins]
        if result:
            return result
        else:
            return []

    def get_origin_by_id(self, oid):
        result = self.mdb.get_origin_by_id(oid)
        return result

    # TAGS
    
    def get_tagged_entities(self, tag_key, tag_value=None):
        props="{key:$key, value:$value}" if tag_value else "{key:$key}"
        ents_by_tag = self.mdb.get_with_statement(
            f"match (t:tag {props}) "
            "with t "
            "match (e)-[:has_tag]->(t) "
            "with t, {type:head(labels(e)), ent:e} as ee "
            "return t.key as tag_key, t.value as tag_value, collect(ee) as entities",
            {"key": tag_key, "value": tag_value})

        for i in range(0, len(ents_by_tag)):
            for j in range(0, len(ents_by_tag[i]['entities'])):
                ents_by_tag[i]['entities'][j]['plural'] = plural[ents_by_tag[i]['entities'][j]['type']]
        return ents_by_tag

    def get_tags_and_values(self):
        tags = self.mdb.get_tags_and_values()
        return tags

    # SEARCH

    def search_entity_handles(self, qstring):
        res = {}
        res["nodes"] = self.mdb.query_index("nodeHandle", qstring) or []
        res["properties"] = self.mdb.query_index("propHandle", qstring) or []
        res["relationships"] = self.mdb.query_index("edgeHandle", qstring) or []
        return res

    def search_terms(self, qstring, search_values=True,
                     search_definitions=True):
        return self.mdb.search_terms(qstring, search_values=search_values,
                                     search_definitions=search_definitions)



import sys
import re
from .. import mdb
from flask import (
    render_template,
    flash,
    redirect,
    url_for,
    request,
    g,
    jsonify,
    current_app,
    Response,
    abort,
)
from flask_paginate import Pagination, get_page_parameter
from .forms import SearchForm, SelectModelForm, SelectVersionForm
# import app.search
from . import bp
from ..util import get_yaml_for
from ..mdb import mdb
# from app.arc import diff_mdf



@bp.before_app_request
def before_request():
    g.search_form = SearchForm()
    g.locale = 'EN_US'


@bp.route('/', methods=['GET', 'POST'])
@bp.route("/index", methods=["GET", "POST"])
def index():
    return render_template(
        "index.html",
        title="Home",
    )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


@bp.route("/models/<name>")
@bp.route("/models")
def models(name=None):
    # format = request.args.get("format")
    m = mdb()

    if name is not None:
        models_ = m.get_model_by_name(name)  # list of versions of model <name>
        select_form = SelectVersionForm()
        select_form.version.choices = current_app.config["VERSIONS_BY_MODEL"][name]
        return render_template(
            "mdb-model.html",
            title="Model: {}".format(models_[0].handle),
            mdb=models_,
            subtype="main.models",
            display="detail",
            form=select_form,
        )

    else:
        models_ = m.get_list_of_models()
        return render_template(
            "mdb-model.html",
            title="Models",
            mdb=sorted(models_, key=lambda x:x["name"]),
            subtype="main.models",
            display="list",
        )

@bp.route("/<entities>", defaults={'id': None}, methods=['GET','POST'])
@bp.route('/<entities>/<id>', methods=['GET','POST'])

def entities(entities, id):

    if request.form.get("format"):
        format = request.args.get("format")
    elif request.form.get("export"):
        format = "json"
    else:
        format = ""

    model = request.args.get("model") or request.form.get("model") or "ALL"
    version = request.args.get("version") or request.form.get("version") or None
    id_ = request.args.get("id")
    page = request.args.get(get_page_parameter(), type=int, default=1)
    select_form = SelectModelForm()
    select_form.model.choices = [(x, x) for x
                                 in current_app.config["MODEL_LIST"]]
    if model != "ALL":
        select_form.version.choices = [(x, x) for x
                                       in current_app.config["VERSIONS_BY_MODEL"][model]]
        select_form.version.choices.insert(0, ("ALL", "ALL"))
    else:
        select_form.version.choices = [("ALL", "ALL")]
        version = "*"
        
    current_app.logger.info(f"model: {model}, version: {version}, page: {page}, format: {format}")
    id = id or id_
    m = mdb()
    dispatch = {
        "nodes": {
            "title": "Node",
            "list_title": "Nodes",
            "template": "mdb-node.html",
            "subtype": "nodes",
            "sort_key": lambda x:(x[1], x[2], x[3]),
            "get_by_id": m.get_node_by_id,
            "get_list":m.get_list_of_nodes,
        },
        "properties": {
            "title": "Property",
            "list_title": "Properties",
            "template": "mdb-property.html",
            "subtype": "properties",
            "sort_key": lambda x: (x["prop_handle"], x["node_model"], x["node_handle"]),
            "display": "prop-tuple",
            "get_by_id": m.get_property_by_id,
            "get_list": m.get_list_of_properties,
        },
        "valuesets": {
            "title": "Value Set",
            "list_title": "Value Sets",
            "template": "mdb-valueset.html",
            "subtype": "valuesets",
            "sort_key": lambda x: x["handle"],
            "display": "list",
            "get_by_id": m.get_valueset_by_id,
            "get_list": m.get_list_of_valuesets,
        },
        "terms": {
            "title": "Term",
            "list_title": "Terms",
            "template": "mdb-term.html",
            "subtype": "terms",
            "sort_key": lambda x: (x["value"].lower()  if type(x["value"]) == str else ""),
            "display": "term-tuple",
            "get_by_id": m.get_term_by_id,
            "get_list": m.get_list_of_terms,
        },
        "origins": {
            "title": "Origin",
            "list_title": "Origins",
            "template": "mdb-origin.html",
            "subtype": "origins",
            "sort_key": lambda x:list(x.values())[0],
            "display": "list",
            "get_list": m.get_list_of_origins,
            "get_by_id": m.get_origin_by_id
        },
    }

    # A: single entity
    if id is not None:
        ent_ = dispatch[entities]["get_by_id"](id)
        if ent_ is None or not bool(ent_):
            return render_template('/errors/400.html'), 400
        if type(ent_) == dict and ent_.get("has_properties"): # nodes only kludge
            ent_['has_properties'].sort(key=lambda x:x['handle'])
        if format == "json":
            return jsonify(ent_)
        else:
            return render_template(
                dispatch[entities]["template"],
                title=dispatch[entities]["title"],
                mdb=ent_,
                subtype=dispatch[entities]["subtype"],
                display="detail",
            )

    # B: list, filter by model
    get_list_args = (model, version)
    if (entities in ['origins', 'terms']):
        get_list_args = ()
    ents_ = dispatch[entities]["get_list"](*get_list_args)

    if format == "json":
        return jsonify(ents_)
    else:
        pgurl = url_for('main.entities',entities=entities)
        pgurl += "?page={0}"
        if model:
            pgurl += "&model={}".format(model)
        pagination = Pagination(
            page=page,
            total=len(ents_),
            record_name=entities,
            href=pgurl
            )
        rendered = render_template(
            "mdb.html",
            title=dispatch[entities]["list_title"],
            mdb=sorted(ents_,
                       key=dispatch[entities]["sort_key"]),
            subtype=dispatch[entities]["subtype"],
            display=dispatch[entities].get("display") or "tuple",
            first=(pagination.page-1)*pagination.per_page,
            last=min((pagination.page)*pagination.per_page,len(ents_)),
            pagination=pagination,
            form=select_form,
            model=model,
        )
        # get a load of THIS kludge, dude.
        if model:
            rendered = re.sub('option value="{}"'.format(model),
                              'option selected="true" value={}'.format(model),
                              rendered)

        return rendered
# ---------------------------------------------------------------------------


@bp.route("/terms", defaults={'start': 0}, methods=['GET', 'POST'])
@bp.route("/terms/batch/<start>", methods=['GET', 'POST'])
def terms(start, num=15):
    page = request.args.get(get_page_parameter(), type=int, default=1)
    m = mdb() 
    (batches, tabnames) = m.get_term_batch_info(num)
    activetab = -1
    activesubtab = -1
    subbatches = None
    subtabnames = None
    batch = None
    paging = {}
    if start is not None:
        start = int(start)
        bsize = batches[0]['last'] - batches[0]['first'] + 1
        for i in range(0, len(batches)):
            if batches[i]['first'] <= start and start <= batches[i]['last']:
                activetab = i
                break
        (subbatches, subtabnames) = m.get_term_batch_info(
            num, batches[i]['first'], bsize
        )
        for j in range(0, len(subbatches)):
            if subbatches[j]['first'] <= start and start < subbatches[j]['last']:
                activesubtab = j
                break
        sbsize = subbatches[j]['last'] - subbatches[j]['first'] + 1 
        batch =  mdb.get_term_batch(start, sbsize)
        pgurl = url_for('main.terms',start=start)
        pgurl += "?page={0}"
        pagination = Pagination(
            page=page,
            total=len(batch),
            record_name=entities,
            href=pgurl
        )
        paging["pagination"] = pagination
        paging["first"] = (pagination.page-1)*pagination.per_page
        paging["last"] = min((pagination.page)*pagination.per_page,len(batch))
            
    return render_template(
        "mdb-term-tabs.html",
        title="Terms",
        tabnames=tabnames,
        subtabnames=subtabnames,
        batches=batches,
        subbatches=subbatches,
        activetab=activetab,
        activesubtab=activesubtab,
        batch=batch,
        paging=paging,
    )


@bp.route("/tags", defaults={'key':None,'value':None},methods=['GET','POST'])
@bp.route("/tags/<key>", methods=['GET','POST'], defaults={'value':None})
@bp.route("/tags/<key>/<value>", methods=['GET','POST'])
def tags(key=None,value=None,model=None):
    key = key or request.args.get("key")
    val = value or request.args.get("value")
    model = model or request.args.get("model")
    ents = []
    m = mdb()
    format = request.args.get("format")
    if request.form.get("format"):
        format = request.args.get("format")
    elif request.form.get("export"):
        format = "json"
    else:
        pass

    if key:
        ents = m.get_tagged_entities(key, val)
    else:
        ents = m.get_tags_and_values()

    if format == "json":
        return jsonify(ents)
    else:
        if key:
            return render_template(
                "mdb-tag.html",
                title="Tagged Entities",
                key=key,
                value=val,
                ents=ents,
                display='entities')
        else:
            return render_template(
                "mdb-tag.html",
                title="Tags",
                ents=ents,
                display='tags')


@bp.route("/search")
def search():
    if not g.search_form.validate():
        return redirect(url_for("main.index"))
    m = mdb()
    format = request.args.get("format")
    if request.form.get("format"):
        format = request.args.get("format")
    elif request.form.get("export"):
        format = "json"
    else:
        pass

    qstring = request.args.get("qstring")
    entdisplay = request.args.get("entdisplay") or "nodes"
    ents = []
    thing = ""
    if request.args.get("terms"):
        ents = m.search_terms(qstring)
        thing = "terms"
    elif request.args.get("models"):
        ents = m.search_entity_handles(qstring)
        thing = "models"
    else:
        abort(400)
    if format == 'json':
        return jsonify(ents)
    paging = {}
    pg_tot = 0
    if not ents:
        thing = "no_hits"
    if thing == "terms":
        pg_tot = len(ents)
        pagination = Pagination(
            page=request.args.get("page", 1, type=int),
            record_name="Hits",
            total=pg_tot,
            per_page=current_app.config["HITS_PER_PAGE"],
        )
        paging["terms"] = {"pagination": pagination}
        paging["terms"]['first'] = (pagination.page-1)*pagination.per_page
        paging["terms"]['last'] = min(pagination.page*pagination.per_page, len(ents))
        entdisplay = "terms"
    elif thing == "models":
        paging = None
        # for ent in ("nodes", "properties", "relationships"):
        #     pagination = Pagination(
        #         page=request.args.get("page",1,type=int),
        #         record_name="Hits",
        #         total=len(ents[ent]),
        #         anchor=ent,
        #         per_page=current_app.config["HITS_PER_PAGE"],
        #     )
        #     paging[ent] = {"pagination": pagination}
        #     paging[ent]['first'] = (pagination.page-1)*pagination.per_page
        #     paging[ent]['last'] = min(pagination.page*pagination.per_page,len(ents[ent]))
    
    return render_template(
        "search.html",
        title="Search",
        ents=ents,
        thing=thing,
        q=qstring,
        entdisplay=entdisplay,
        paging=paging,
    )

@bp.route("/about-mdb")

def about_mdb():
    return render_template("about-mdb.html", title="About MDB")


@bp.route("/about-sts")

def about_sts():
    return render_template("about-sts.html", title="About STS")

@bp.errorhandler(413)
def too_large(e):
    return "File is too large", 413

@bp.route('/reports', methods=['GET', 'POST'])
def reports():
    return render_template('reports.html',)

@bp.route('/versionhistory', methods=['GET', 'POST'])
def versionhistory():
    return render_template('version-history.html')
